#include "configs.h"
#include "daem.h"
#include "helpers.h"
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include "config_struct.h"

char *makePathRelative(const char *path, const char *relativeto){
    if(path[0]=='/'){
        return (char*)path;
    } else{
        int reltolen = strlen(relativeto);
        int pathlen = strlen(path);
        // check if adding a slash to the relative path is necessary
        int addslash = (relativeto[reltolen - 1] != '/') ? 1 : 0;
        char * ret = malloc(reltolen + pathlen + addslash + 1);

        strcpy(ret, relativeto);
        if(addslash) ret[reltolen] = '/';
        strcpy(ret + reltolen + addslash, path);

        return ret;
    }
}

int strtoim_wdef(char *str, int min, int max, int def, int *err){
    errno=0;
    int r = strtol(str, NULL, 10);
    if (errno) {
        if(err)*err=1;
        return def;
    }

    if (r>max) return max;
    if(r<min) return min;

    return r;
}


int setValueCfg(char *property, char *value, void *c){
    struct Config *cfg = c;

    if(!property || !value) return 0;
    const char *const infiledefs[] = {
        "loopInterval",
        "logLevel",
        "selfLog",
        "logdbdir",
        "daemondb",
        "seperateLogs",
    };

    int v=-1;

    for (int i=0; i<sizeof(infiledefs)/sizeof(infiledefs[0]); i++) {
        if(strcmp(infiledefs[i], property)==0){
            v=i;
        }
    }
    switch (v) {
        case 0:{
            cfg->loopinterval = strtoim_wdef(value, 0, 100000, 500, NULL);
            return 0;
        }
        case 1:{
            cfg->loglevel = strtoim_wdef(value, 0, 4, 0, NULL);
            return 0;
        }
        case 2:{
            cfg->selflog = value;
            return 1;
        }
        case 3:{
            cfg->logdir = value;
            return 1;
        }
        case 4:
            cfg->daemdir = value;
            return 1;
        case 5:{
            cfg->seperatelogs = strcmp(value, "true") == 0;
            return 0;
        }
        default:{
            fprintf(stderr, "property: \"%s\" does not exist\n", property);
            return 0;
        }
    
    }
}

int setValueDaemon(char *property, char *value, void* d){
    struct Daemon *daemon = d;

    if(!property || !value) return 0;
    const char* const infiledefs[] = {
        "exec",
        "argv",
        "logf",
        "errf",
        "restartcrash",
        "restartexit",
        "restartinterval"
    };

    int v=-1;

    for (int i=0; i<sizeof(infiledefs)/sizeof(infiledefs[0]); i++) {
        if(strcmp(infiledefs[i], property)==0){
            v=i;
        }
    }

    switch (v) {
        case 0:{ // exec
            daemon->exec = value;
            break;
        }
        case 1:{ // argv
            daemon->argv = value;
            break;
        }
        case 2:{ // logf
            daemon->logf = makePathRelative(value, cfgptr->logdir);
            break;
        }
        case 3:{ // errf
            daemon->errf = makePathRelative(value, cfgptr->logdir);
            break;
        }
        case 4:{ // restartcrash
            if(strcmp(value, "true")==0){
                daemon->configflags|=RESTART_ON_CRASH;
            }
            return 0;
        }
        case 5:{ // restartexit
            if(strcmp(value, "true")==0){
                daemon->configflags|=RESTART_ON_EXIT;
            }
            return 0;
        }
        case 6:{ // restartinterval
            daemon->restartinterval = strtol(value, NULL, 10);
            return 0;
        }

        default:{
            fprintf(stderr, "property: \"%s\" does not exist\n", property);
            return 0;
        }
    }
    return 1;
}

void parseCfg(char*cf, int (setVal)(char*,char*,void*), void *v){
    char* c = cf;

    char* property=c;
    char* value=0;

    // property name length and value length
    int pl=0, vl=0;

    // space counter
    int spctr = 0;

    //0:readproperty
    //1:ignorespace
    //2:readvalue
    //3:gotonextline
    int state=0;
    
    int i=0;

    for (c=cf; *c != '\0'; c++) {

        if(*c=='\n' || *c=='\0'){ 
            //newline new property
            state=0;
            //process read property and value
            if(property && value){
                // remove trailing spaces
                vl -= spctr;

                // property char string
                char pc[pl+1];
                strncpy(pc, property, pl);
                pc[pl]=0;

                // value char string
                char *vc = malloc(sizeof(char)*(vl+1));
                strncpy(vc, value, vl);
                vc[vl]=0;

                if(!setVal(pc, vc, v)){
                    free(vc);
                }
            }

            vl=pl=0;
            property=c+1;
            value=0;
            continue;
        } else if(state==3){
            //comment, skip everything
            continue;
        } else if(*c=='#'){
            //start comment
            state=3;
            continue;
        } else if(*c=='=' && state==0){
            //switch from property to value
            state=1;
            continue;
        } else if(*c==' ' && (state == 0 || state==1)){
            continue;
        } else if(state==0){    
            pl++;
        } else if(state==2){
            vl++;
        } if(state == 1 && *c!=' '){
            value=c;
            state=2;
            vl++;
        }
        // increment space counter
        if(*c==' ') spctr++;
        else spctr = 0;
    }
}

char *getNameFromPath(const char *path){
    char *c = (char*)path;
    char *s;
    while(*c != 0){
        if(*c=='/') s = c;
        c++;
    }
    s++;
    int slen = strlen(s);
    slen -= strlen(".daem.conf");
    if(slen<=0) return 0;
    char *ret = malloc(slen * sizeof(char) + 1);
    memcpy(ret, s, slen);
    ret[slen] = 0;
    return ret;
}

// note: path may be modified
struct Daemon readDaemonCfg(const char *path, int *err){
    struct Daemon d = {0};

    // read File
    if(loglevel > 3) printf("Reading daemon config %s ...\n", path);
    char* cf = readFileC(path);
    if (!cf){
        if(err)*err=1;
        return d;
    }

    // parse the config
    parseCfg(cf, setValueDaemon, &d);

    // set the name
    d.name = getNameFromPath(path);
    if(!d.name){
        if(err)*err=1;
        return d;
    }

    // setup defaults
    if(!d.logf) {
        char *cs = concatenateString(d.name, ".log");
        d.logf = makePathRelative(cs, cfgptr->logdir);
        free(cs);
    }
    if(!d.argv) d.argv = copyOfString("");

    // set errf to logf if not defined
    if(!d.errf){
        d.errf = copyOfString(d.logf);
    }

    // clean up unused variables
    free(cf);
    

    // check for completeness and add the name of the daemon
    d.configflags |= ISCOMPLETE;

    if(!d.exec){
        d.configflags &=~ISCOMPLETE;
        if(err)*err=2;
        if(loglevel > 0) fprintf(stderr, "Could not find exec of Daemon.\n");
    }

    return d;
}

struct Config readSelfCfg(const char *path, int *err){
    // since selfconfig is not freed or modified don't use copyOfString
    struct Config sc={};
    sc.loopinterval = 500;
    sc.loglevel = 0;
    sc.selflog = "log/udaem.log";
    sc.logdir = "log";
    sc.shell = "/bin/sh"; // TODO: implement overriding this var
    sc.seperatelogs = 0;

    char *cf = readFileC(path);
    if(!cf){
        if(err)*err=1;
        return sc;
    }

    parseCfg(cf, setValueCfg, &sc);
    if(!sc.daemdir) *err = 2;
    
    free(cf);
    return sc;
}

struct Daemon mergeDaemonToRunning(struct Daemon running, struct Daemon free){
    running.name = free.name;
    running.exec = free.exec;
    running.argv = free.argv;
    running.errf = free.errf;
    running.logf = free.logf;
    running.configflags = free.configflags;
    running.restartinterval = free.restartinterval;
    running.state |= shouldRestartDaemon(free, running);
    return running;
}

struct Daemon mergeDaemonToFree(struct Daemon free, struct Daemon running){
    free.state = running.state | shouldRestartDaemon(free, running);
    free.pid = running.pid;
    free.exitstatus = running.exitstatus;
    free.exitreason = running.exitreason;
    return free;
}

int shouldRestartDaemon(struct Daemon a, struct Daemon b){
    int rec=0;
    int req=0;

    rec += !isIdenticalString(a.name, b.name);
    rec += !isIdenticalString(a.exec, b.exec);
    rec += !isIdenticalString(a.argv, b.argv);

    // logfile changes only recomend restart
    // TODO: think about restartrequired and if it is really necessary
    rec += !isIdenticalString(a.errf, b.logf);
    rec += !isIdenticalString(a.logf, b.logf);

    if(req>0) return RESTARTREQUIRED;
    if(rec>0) return RESTARTRECOMEND;
    return 0;
}